/* M2P  Maple-to-POV-Ray v0.9.5b  
 * By Bjrn Vaggen Konestabo for the Institute of Mathematics, University of
 * Oslo.
 *
 * Note:
 *   This program was made to give quick results. 
 *   Code optimization has NOT been a priority.
 */

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <float.h>

#define max(A,B) ((A)>(B) ? (A):(B))
#define min(A,B) ((A)<(B) ? (A):(B))

/* Internal message codes */
#define SB_OVERFLOW "M2P:Symbol Buffer Overflow"
#define MESH        0
#define GRID        1

#define COLOUR      1
#define HUE         2
#define RGB         4
#define MULTIPLE    8
#define NDEF    10000

/* ARGUMENTS */
#define SMOOTH           1
#define NORMAL           2
#define TRION            4
#define CENTER_RESCALE   8
#define SPHERE          16
#define PATCH           32
#define LINE            64
#define BACK           128
#define CONTOUR        256
#define BOX            512
#define VERBOSE       1024
#define WHITEBACK     2048

#define DEFAULT_ASPECT   1.33

#define FEEDBACK  if(arg&VERBOSE)

#ifndef PI 
#define PI    3.1415926535897932
#endif

#ifndef SEEK_SET
#define SEEK_SET 0 
#endif

struct ve{
  double x;
  double y;
  double z;
};
struct po{
  double x;
  double y;
  double z;
};
struct col{
  double r;
  double g;
  double b;
};

/* Prototyping is good for you */
int mesh(FILE *in,FILE *out, int arg, int otype);
void camera(double dist,double aspect, FILE *out);
struct ve *cross_prod(struct ve *v1, struct ve *v2, struct ve *ret);
struct ve *vector(struct po *p1, struct po *p2, struct ve *ret);
double angle(struct ve *v1, struct ve *v2);
int eq(double a, double b);
int peq(struct po *p1, struct po *p2);
int arguments();
char *sread(FILE *in);
double nread(FILE *in);
double num_read(char *in); 
void center_rescale(FILE *out); 
void hue(double pos, struct col *c); 
void ve_add(struct ve *v1, struct ve *v2);
struct po *po_add(struct po *p1, struct ve *v1, double factor,struct po *ret);
struct po *po_sub(struct po *p1, struct ve *v1, double factor,struct po *ret);
struct po *po_avrg(struct po *p1, struct po *p2, 
		  struct po *p3, struct po *p4, struct po *ret);

/* POV-Ray funcs */
void s_triangle(struct po *p1, struct po *p2, struct po *p3, 
		struct ve *n1, struct ve *n2, struct ve *n3, struct col *c, FILE *out);
void cylinder(struct po *p1, struct po *p2, double r, FILE *out);
void sphere(struct po *p, double r, FILE *out);
void triangle(struct po *p1, struct po *p2, struct po *p3, struct col *c, FILE *out);


 struct po max={-100000,-100000,-100000};
 struct po min={100000,100000,100000};  /* Used for centering */

 struct po lmax={-100000,-100000,-100000};
 struct po lmin={100000,100000,100000};  /* Used for centering */

 double temp;    
 double amax=40;      /* The angle threshold. If the angle between two
		       * triangles is greater than the amax setting,
                       * It is assumed that the surface is not smooth */

int
main(int argc, char **argv)
{
  FILE *fo,*fp;
   char *symbol,*fn;
  int c,i,j,ok=1;
  int arg;
  int object_open=0;
  double theta,phi;
  double aspect=DEFAULT_ASPECT;

  fn=(char *) malloc(256);  /* filename */
  arg=(TRION+CENTER_RESCALE);   /* Default switches */
  fn[0]=0;

  /* Checking for swiches */

  for(i=1;i<argc;i++){
    if (argv[i][0]=='-') {
      j=0; while((c=argv[i][++j])!=0) {
	if (isdigit(c) || c=='.') continue;
	if(c=='s') arg|=SMOOTH;
	else if(c=='n') arg|=NORMAL;
	else if(c=='h') arg^=TRION;
	else if(c=='o') arg^=CENTER_RESCALE;
	else if(c=='b') arg|=SPHERE;
	else if(c=='p') arg|=PATCH;
	else if(c=='l') arg|=LINE;
	else if(c=='c') arg|=CONTOUR;
	else if(c=='e') arg|=BOX;
	else if(c=='B') arg|=BACK;
	else if(c=='W') arg|=WHITEBACK;
	else if(c=='V') arg|=VERBOSE;
	else if(c=='A') {
	  aspect=num_read(&argv[i][j+1]);
	  if (aspect==DBL_MAX) {
	    printf("Corrupt command line. Setting Aspect ratio to default.\n");
	    aspect=DEFAULT_ASPECT;
	  }
	}
	else FEEDBACK printf(" Switch '%c' ignored\n",c);
      }
    }
    else if (fn[0]==0) strcpy(fn,argv[i]);
  }
  if (!fn[0]) return arguments(argv[0]);
	  
  /* Filenames */
  fp=fopen(fn,"r");
  strcat(fn,".pov");
  fo=fopen(fn,"wb");

  fprintf(fo,"#declare LOWER_GRAD_COL=color rgb <.1, .1, .3>\n");
  fprintf(fo,"#declare UPPER_GRAD_COL=color rgb <.3, .3,  1>\n"); 
  fprintf(fo,"#include \"m2p.inc\"\n");
 
  /* Can't forget the camera */
  camera(-6,aspect,fo); 
  
  if (arg&BACK) fprintf(fo,"object{BACKGROUND}\n");
  if (arg&WHITEBACK) fprintf(fo,"background{color rgb <1,1,1>}\n");
  /* Start readin' symbols */

  theta=45; 
  phi=45;

  fprintf(fo,"union{\n");

  while((symbol=sread(fp))!=0 && ok) { 
    if (strcmp("MESH",symbol)==0) {ok=mesh(fp,fo,arg,MESH); continue;}
    if (strcmp("GRID",symbol)==0) {ok=mesh(fp,fo,arg,GRID); continue;}

    /* A new object */
    if (strcmp("PLOT3D",symbol)==0){ 
      if (object_open) {
	fprintf(fo,"texture{Texture1}\n");
	if (arg&CONTOUR) 
	  fprintf(fo,"texture{Contour scale %f}\n",(lmax.y-lmin.y)/20.0);
	fprintf(fo,"}\n");
      }
      lmax.x=lmax.y=lmax.z=-100000;
      lmin.x=lmin.y=lmin.z=100000;
      object_open=1;
      fprintf(fo,"union{\n");
      continue;
    } 

    /* Reads camera positions */
    if (strcmp("ORIENTATION",symbol)==0) {
      theta=nread(fp);
      phi=nread(fp);
      continue;
    }
    /* Check for style */
    if (strcmp("STYLE",symbol)==0) {
      symbol=sread(fp);
      if (strcmp("PATCHCONTOUR",symbol)==0) arg|=CONTOUR;
      if (strcmp("CONTOUR",symbol)==0) arg|=CONTOUR;
      continue;
    }
    if (strcmp(SB_OVERFLOW,symbol)==0) ok=0; 
  }

  /* If something went horribly wrong */
  if (!ok) {
    FEEDBACK printf("M2P: File is corrupt.\n");
    fclose(fo); fclose(fp);
    return -2;
  }

  if (object_open) {
    fprintf(fo,"texture{Texture1}\n");
    if (arg&CONTOUR) 
      fprintf(fo,"texture{Contour scale %f}\n",(lmax.y-lmin.y)/20.0);
    fprintf(fo,"}\n");
  }

  if (arg&CENTER_RESCALE) center_rescale(fo);
  fprintf(fo,"rotate <0,%-.5f,0>\n",(90.0-theta+180));
  fprintf(fo,"rotate <%-.5f,0,0>\n",(phi-90.0));
  fprintf(fo,"}\n");

  if (arg&BOX) 
    fprintf(fo,"object{BOX rotate <0,%-.5f,0> rotate <%-.5f,0,0>}\n");

  fprintf(fo,"object{Light1}\n");  
  fclose(fo); fclose(fp);
  printf("Done.\n");
  return 0;
}


/* ======================================================================= */


int
mesh(FILE *in,FILE *out,int arg, int otype) 
{
  int siz[3],count,i,j,c,nline,lpnt;
  int colour=0;
  double fromx,tox,fromz,toz;
  long tfp;   

  FEEDBACK printf("-- START of new object.\n");
  FEEDBACK printf("   Scanning object.\n");

  if (otype==GRID) {
    fromx=-nread(in);
    fgetc(in);
    fgetc(in);
    tox=-nread(in);
    fromz=nread(in);
    fgetc(in);
    fgetc(in);
    toz=nread(in);
  }

  siz[0]=siz[1]=siz[2]=0;
  tfp=ftell(in);      /* Remember fileposition */

  /* The mesh is divided into rows of points. */
  /* This routine counts the number of rows and points */
  i=0; j=1;
  while (siz[0]<1){
    c=getc(in);
    if (c=='[') i++;
    if (c==',') j++;
    if (i>3 || i==0) return 0;
    if (c==']') siz[--i]++;
  }

  if (otype==GRID) siz[2]=j;

  nline=siz[1];        /* Number of rows */
  lpnt=siz[2]/siz[1];  /* Points per row */

  /* Checking for colors */
  if (strcmp(sread(in),"COLOR")==0) {
    colour=COLOUR;
    if (strcmp(sread(in),"HUE")==0) colour|=HUE;
    else colour|=RGB;

    /* Count the commas in the COLOR() statement */
    i=0; j=0;
    while ((c=getc(in))!=EOF && c!=')'){
      if (c==',') j++;
    }
    /* The object is not multicoloured */
    if (j>3) colour|=MULTIPLE;
  }
  
  /* Trace back to start of object */
  fseek(in,tfp,SEEK_SET);
  count=0;
  
  /* NEW BLOCK  STARTS HERE */
  {
    int i,j,r;
    int fj,fi; 

    double x,y,z;
    double linewidth=0;
    struct col clr={NDEF,NDEF,NDEF};
    struct col scol={NDEF,NDEF,NDEF}; /* single color */

    struct col **cl; /* Colors */
    struct po **msh; /* All the coordinates for the mesh */
    struct ve **sn; /* The normal vector for all mesh squares*/
 
    /* struct col cl[nline][lpnt];  */ /* Colors */
    /* struct po msh[nline][lpnt];*/ /* All the coordinates for the mesh */
    /* struct ve sn[nline-1][lpnt-1];*/ /* The normal vector for all mesh squares*/
    struct ve tmpvec1,tmpvec2;       /* Temp values used to hold data from */
    struct po tmppo1,tmppo2,tmppo3;   /* various function calls */

    cl =  malloc (nline*sizeof (struct col *));
    msh= malloc (nline*sizeof (struct po *));
    sn = malloc ((nline-1)*sizeof (struct ve *));


    for (i=0; i<nline;i++){
        cl[i] = (struct col *) malloc (lpnt*sizeof (struct col));
        msh[i]= (struct po *)  malloc (lpnt*sizeof (struct po));
        if (i) sn[i-1] = (struct ve *)  malloc (lpnt*sizeof (struct ve));
    }


    FEEDBACK printf("   Reading: \n");

    if (otype==GRID) {
      double stepx,stepz,px,pz;
 
      /* The Object type is NOT a MESH, 
       * create the additional coordinates 
       * to convert the GRID into a MESH */

      lmax.x=max(fromx,tox); lmin.x=min(fromx,tox);
      lmax.z=max(fromz,toz); lmin.z=min(fromz,toz);

      stepx=(tox-fromx)/(nline-1);
      stepz=(toz-fromz)/(lpnt-1);

      px=fromx;
      for(r=0;r<nline;r++) {
	pz=fromz;
	for(j=0;j<lpnt;j++) {
	  x=msh[r][j].x=px;
	  z=msh[r][j].z=pz;
	  y=msh[r][j].y=nread(in);
	  if (y==DBL_MAX) return 0;
	  lmax.y=max(y,lmax.y);
	  lmin.y=min(y,lmin.y);
	  pz+=stepz;
        }
	px+=stepx;
      }
    }
    else {
      /* The object is MESH */
      for(r=0;r<nline;r++) {
	/* Read one row of points */
	for(j=0;j<lpnt;j++) {
	  x=msh[r][j].x=-nread(in);
	  z=msh[r][j].z=nread(in);
	  y=msh[r][j].y=nread(in);
	  if (x==DBL_MAX || y==DBL_MAX || z==DBL_MAX) return 0;
	  lmax.x=max(x,lmax.x); lmax.y=max(y,lmax.y); lmax.z=max(z,lmax.z);
	  lmin.x=min(x,lmin.x); lmin.y=min(y,lmin.y); lmin.z=min(z,lmin.z);
	}
      }
    }
    max.x=max(max.x,lmax.x); max.y=max(max.y,lmax.y); max.z=max(max.z,lmax.z);
    min.x=min(min.x,lmin.x); min.y=min(min.y,lmin.y); min.z=min(min.z,lmin.z);

/* Adding colours. (This is experimental!) */
    if (colour){
      if (colour&MULTIPLE) {
	sread(in); sread(in);
	if (colour&HUE) {
	  for(r=0;r<nline;r++) {	
	    for(j=0;j<lpnt;j++) {
	      x=nread(in);
	      hue(x,&cl[r][j]);
	    }
	  }
	}
	else if (colour&RGB) {
 	  for(r=0;r<nline;r++) { 
	    for(j=0;j<lpnt;j++) {		
	      cl[r][j].r=nread(in);
	      cl[r][j].g=nread(in);
	      cl[r][j].b=nread(in);
	    }
	  }
	}
      }
      else {
	if (colour&HUE) hue(nread(in),&scol); 
	else {
	  scol.r=nread(in);
	  scol.g=nread(in);
	  scol.b=nread(in);
	}
      }
    }
    for(r=0;r<nline-1;r++) {
      for(j=0;j<lpnt-1;j++) {
	count++;
	x=msh[r][j].x-msh[r+1][j+1].x;
	y=msh[r][j].y-msh[r+1][j+1].y;
	z=msh[r][j].z-msh[r+1][j+1].z;
	linewidth+=sqrt(x*x+y*y+z*z);
      }
    }

    linewidth/=(double)(count*30);

    FEEDBACK printf("   %d rows of %d points each.  %d points total.\n",
	   nline,lpnt,nline*lpnt);

    /* Time to mesh some triangles */
    FEEDBACK printf("   -- Creating object:\n");

    /* Find the mesh square normals */
    FEEDBACK printf("      Calculating mesh square normals.\n");      
    {
      int a,b;
      for(i=0;i<nline-1;i++) {
	fi=i+1; a=i;
	for(j=0;j<lpnt-1;j++) {
	  fj=j+1; b=j;
	  
	  /* To make sure the object it tries to calculate the normal 
	     From is indeed a triangle and NOT a line (oops) */
	  if (peq(&msh[i][j],&msh[i][fj])) a=fi;
	  if (peq(&msh[i][j],&msh[fi][j])) b=fj;
	  
	  cross_prod(vector(&msh[i][j],&msh[fi][b],&tmpvec1),
		     vector(&msh[i][j],&msh[a][fj],&tmpvec2),&sn[i][j]);
	}
      }
    }

    /* Create the cylinders representing the mesh normals */
    if (arg&NORMAL) {
      int fi,fj; 
      for(i=0;i<nline-1;i++) { 
	fi=i+1; 
	for(j=0;j<lpnt-1;j++) {
	  fj=j+1;  
	  cylinder(po_add(po_avrg(&msh[i][j],&msh[fi][j],
				  &msh[i][fj],&msh[fi][fj],&tmppo3),
			  &sn[i][j],linewidth*15,&tmppo1),
		   po_sub(&tmppo3,&sn[i][j],linewidth*15,&tmppo2),
		   linewidth*2.5,out);
	}
      }
      FEEDBACK printf("      %d cylinder normals created.\n",nline*lpnt);
    }
    
    /* Create the spheres representing the mesh points */
    if (arg&SPHERE) {
      for(i=0;i<nline;i++) for(j=0;j<lpnt;j++) sphere(&msh[i][j],linewidth*10,out);;
      FEEDBACK printf("      %d sphere points created.\n",nline*lpnt);
    }

    /* Create the cylinders representing the mesh edges */
    count=0;
    if (arg&LINE) {
      fprintf(out,"union {\n");
      for(i=0;i<nline;i++) {
	for(j=0;j<lpnt;j++) {
	  if (j<lpnt-1) cylinder(&msh[i][j],&msh[i][j+1],linewidth,out);
	  if (i<nline-1) cylinder(&msh[i][j],&msh[i+1][j],linewidth,out);
	  sphere(&msh[i][j],linewidth,out);
	  count++;
	}
      }
      fprintf(out,"no_shadow\npigment{color rgb <.2,.2,.2>}\n}\n");
      FEEDBACK printf("      %d cylinder edges created.\n",count);
    }
    
    count=0;    
    if (arg&TRION && !(arg&PATCH)) {
      int t,nr,map;
      int ii[2],jj[2],t1,t2;
      static int se[4][2]={{0,0},{0,1},{1,1},{1,0}};
      struct ve pnr[4];  /* Smooth triangle normals */
      clr.r=clr.g=clr.b=NDEF;  /* Colour blank */

      fprintf(out,"union{\n");
      if (arg&SMOOTH) {
 	for(i=0;i<nline-1;i++) {   
	  ii[1]=i+1;
	  ii[0]=(i-1+nline-1)%(nline-1);
	  for (j=0;j<lpnt-1;j++) {
	    jj[1]=j+1;
	    jj[0]=(j-1+lpnt-1)%(lpnt-1);
	    
	    /* Ugly, very ugly. Calculates the correct point normals. */
	    for(t=0;t<4;t++) {
	      map=1+2+4+8;
	      t1=ii[se[t][0]]%(nline-1);	
	      t2=jj[se[t][1]]%(lpnt-1);
	      
              /* Checks if the mesh is connected at the ends */
	      if (j==0 && se[t][1]==0 && 
		  !peq(&msh[i][j],&msh[i][lpnt-1])) map&=1+8;
	      
	      if (i==0 && se[t][0]==0 && 
		  !peq(&msh[i][j],&msh[nline-1][j])) map&=1+2;
	      
	      if (j==lpnt-2 && se[t][1]==1 && 
		  !peq(&msh[i][t2],&msh[i][lpnt-1])) map&=1+8;
	      
	      if (i==nline-2 && se[t][0]==1 && 
		  !peq(&msh[t1][j],&msh[nline-1][j])) map&=1+2;
	
	      /* Break the surface if the angle is too great. */ 
      
	      if (map&8 && angle(&sn[i][j],&sn[t1][j])>amax) map&=1+2+4;
	      if (map&4 && angle(&sn[i][j],&sn[t1][t2])>amax) map&=1+2+8;
	      if (map&2 && angle(&sn[i][j],&sn[i][t2])>amax) map&=1+4+8;
	      
	      /* Find the average normals */
	      pnr[t]=sn[i][j]; 
	      if (map&2) ve_add(&pnr[t],&sn[i][t2]);
	      if (map&4) ve_add(&pnr[t],&sn[t1][t2]);
	      if (map&8) ve_add(&pnr[t],&sn[t1][j]);

      	      nr=1+((map&2)/2)+((map&4)/4)+((map&4)/8);

	      pnr[t].x/=nr; pnr[t].y/=nr; pnr[t].z/=nr;
	    }
	    
	    if (colour&MULTIPLE) {
		clr.r=(cl[i][j].r+cl[i][jj[1]].r+
		       cl[ii[1]][jj[1]].r+cl[ii[1]][j].r)/4;
		clr.g=(cl[i][j].g+cl[i][jj[1]].g+
		       cl[ii[1]][jj[1]].g+cl[ii[1]][j].g)/4;
		clr.b=(cl[i][j].b+cl[i][jj[1]].b+
		       cl[ii[1]][jj[1]].b+cl[ii[1]][j].b)/4;
	    }

	    s_triangle(&msh[i][j],&msh[i][jj[1]],&msh[ii[1]][jj[1]],
		       &pnr[0],&pnr[1],&pnr[2],&clr,out);
	    s_triangle(&msh[i][j],&msh[ii[1]][jj[1]],&msh[ii[1]][j],
		       &pnr[0],&pnr[2],&pnr[3],&clr,out);
	    count+=2;
	  }
	}
      }
      else{
	int fi,i,j; 
	/* Flat shaded triangles */
	for(i=0;i<nline-1;i++) {
	  fi=i+1;
	  for (j=0;j<lpnt-1;j++) {
	    fj=j+1;
	    if (colour&MULTIPLE) {
		clr.r=(cl[i][j].r+cl[i][fj].r+
		       cl[fi][fj].r+cl[fi][j].r)/4;
		clr.g=(cl[i][j].g+cl[i][fj].g+
		       cl[fi][fj].g+cl[fi][j].g)/4;
		clr.b=(cl[i][j].b+cl[i][fj].b+
		       cl[fi][fj].b+cl[fi][j].b)/4;
	    }
   	    triangle(&msh[i][j],&msh[i][fj],&msh[fi][fj],&clr,out);
	    triangle(&msh[i][j],&msh[fi][fj],&msh[fi][j],&clr,out);
	    count+=2;
	  }
	}
      }
      if ((colour&MULTIPLE)==0 && scol.r!=NDEF)
	  fprintf(out,"pigment{color rgb <%f,%f,%f>}\n",scol.r,scol.g,scol.b);
      FEEDBACK printf("      %d triangles created.\n",count);
      fprintf(out,"}\n");
      FEEDBACK printf("   -- END of object creation.\n");
    }
    FEEDBACK printf("-- END of object.\n");

    for (i=0; i<nline;i++){
        free(cl[i]);
        free(msh[i]);
        if (i) free(sn[i-1]);
    }
    free(cl);
    free(msh);
    free(sn);

    return 1;
  }
}

int
arguments(char* com) {
  printf("Usage: %s [-snhbolpceWBV] [-A#.##] file\n\n",com);
  printf("   Switches:\n"
         "      -s  Smooth triangle approximation\n"
         "      -n  Show face normals\n"
         "      -h  Hide faces\n"
         "      -b  Create spheres at all mesh points.\n"
         "      -o  Turn off center and rescale.\n"
         "      -l  Line drawing.\n"
         "      -p  Use bicubic patches. (Not yet implemented)\n"
         "      -c  Topographic contours \n"
         "      -e  Show edge-box\n"
         "      -W  White background (for foils, printouts, etc.)\n"
         "      -B  Background Gradient\n"
         "      -A#.##  Aspect ratio\n"
         "      -V  Verbose messages.\n\n"
         "       Maple-to-POV-Ray v0.9.5a \n");
         
  return -1;
}

/* Finding the angle between two vectors */
double angle(struct ve *v1, struct ve *v2) 
{
  return fabs(acos(((v1->x)*(v2->x)+(v1->y)*(v2->y)+(v1->z)*(v2->z))/
	      (sqrt((v1->x)*(v1->x)+(v1->y)*(v1->y)+(v1->z)*(v1->z))*
               sqrt((v2->x)*(v2->x)+(v2->y)*(v2->y)+(v2->z)*(v2->z))))/PI*180);
}

/* Converting two points into a vector */
struct ve 
*vector(struct po *p1, struct po *p2,struct ve *ret) 
{
  ret->x=p2->x-p1->x;
  ret->y=p2->y-p1->y;
  ret->z=p2->z-p1->z;
  return ret;
}

/* Translating a point by a vector */
struct po
*po_add(struct po *p1, struct ve *v1, double factor, struct po *ret)
{
  ret->x=((p1->x)+((v1->x)*factor));
  ret->y=((p1->y)+((v1->y)*factor));
  ret->z=((p1->z)+((v1->z)*factor));
  return ret;
}

/*  */
struct po
*po_sub(struct po *p1, struct ve *v1, double factor, struct po *ret)
{
  ret->x=((p1->x)-((v1->x)*factor));
  ret->y=((p1->y)-((v1->y)*factor));
  ret->z=((p1->z)-((v1->z)*factor));
  return ret;
}

void
ve_add(struct ve *v1, struct ve *v2)
{
  (v1->x)+=(v2->x);
  (v1->y)+=(v2->y);
  (v1->z)+=(v2->z);
  return;
}

struct po
*po_avrg(struct po *p1, struct po *p2, struct po *p3, 
	struct po *p4, struct po *ret)
{
  ret->x=((p1->x)+(p2->x)+(p3->x)+(p4->x))/4.0;
  ret->y=((p1->y)+(p2->y)+(p3->y)+(p4->y))/4.0;
  ret->z=((p1->z)+(p2->z)+(p3->z)+(p4->z))/4.0;
  return ret;
}

/* Vector cross product  [A,B,C]x[D,E,F]=[BF-CE,CD-AF,AE-BD] */
struct ve
*cross_prod(struct ve *v1, struct ve *v2, struct ve *ret) 
{
  static double x,y,z,l;
  x=((v1->y)*(v2->z)-(v1->z)*(v2->y))*2;
  y=((v1->z)*(v2->x)-(v1->x)*(v2->z))*2;
  z=((v1->x)*(v2->y)-(v1->y)*(v2->x))*2;

  l=sqrt(x*x+y*y+z*z);
  ret->x=(x/l);
  ret->y=(y/l);
  ret->z=(z/l);
  return ret;
}

eq(double a,double b) 
{                 
  if (fabs(a-b)<0.000000001) return 1; 
  return 0;
}

int 
peq(struct po *p1, struct po *p2) {
  if (!eq(p1->x,p2->x)) return 0;
  if (!eq(p1->y,p2->y)) return 0;
  if (!eq(p1->z,p2->z)) return 0;
  return 1;
}

/* Symbol reader */
 char *
sread(FILE *in) 
{
  int c,i;
  static  char symbol[80];
  i=0;
  while ((c=fgetc(in))!=EOF) {
      if (c=='\n') continue;
    if ((c==')' || c==']' || c=='[' || c==',' || c=='(')) {
      if (i) {symbol[i]='\0'; break;}
      continue;
    }
    else symbol[i++]=( char) c;
    if (i==80) return SB_OVERFLOW;
  }
  if (c==EOF) return 0;

  /* printf("%s **\n",symbol);  */
  return symbol;
}
 
/* Number reader */
double
nread(FILE *in) { 
  int c,i;
  static  char number[30];
  i=0;
  while ((c=fgetc(in))!=EOF) {
    if ((c==',' || c==']' || c==' ' || c==')') && i>0) {
      number[i]='\0';
      return strtod(number,0);
    }
    if ((isdigit(c) || c=='.' || c=='-' || c=='+' || c=='e')) {
      number[i++]=( char) c;
      if (i>29) break;
      continue;
    }
  }
  return DBL_MAX;
}



double
num_read(char *in) { 
  int c,i;
  static char number[30];
  i=0;
  for(;;){
    c=*(in++);
    if (isdigit(c) || c=='.' || c=='-' || c=='+' || c=='e') {
      number[i++]=( char) c;
      if (i>29) break;
    }
    else if (i>0) {
      number[i]='\0';
      return strtod(number,0);
    }
  }
  return DBL_MAX;
}



/* Fits the object into a [-1,-1,-1],[1,1,1] cube, without changing the aspect
 * ratio */
void
center_rescale(FILE *out) 
{
  struct po r;
  double fac;

  r.x=(min.x+max.x)/2.0;      
  r.y=(min.y+max.y)/2.0;      
  r.z=(min.z+max.z)/2.0;      
  min.x-=r.x; max.x-=r.x;
  min.y-=r.y; max.y-=r.y;
  min.z-=r.z; max.z-=r.z;
  fac=0;
  
  fac=max(max.x,fac);
  fac=max(max.y,fac);
  fac=max(max.z,fac);
  
  fprintf(out," translate <%-.15f,%-.15f,%-.15f>\n",-r.x,-r.y,-r.z);
  fprintf(out," scale <%-.15f,%-.15f,%-.15f>\n",1/fac,1/fac,1/fac);
}


/* Get the RGB values from a HUE value */
void
hue(double pos, struct col *c) {
  if (pos<=0.166666667) { c->r=1; c->g=pos*6; c->b=0; return; }
  if (pos<=0.333333333) { c->r=1-(pos-0.16666667)*6; c->g=1; c->b=0; return; }
  if (pos<=0.5) { c->r=0; c->g=1; c->b=(pos-0.33333333)*6; return; }
  if (pos<=0.666666667) { c->r=0; c->g=1-(pos-0.5)*6; c->b=1; return; }
  if (pos<=0.833333333) { c->r=(pos-0.666666667)*6; c->g=0; c->b=1; return; }
  c->r=1; c->g=0; c->b=1-(pos-0.833333333)*6;
}

/*
 * ----------------------------------------------------------------------
 *                         The POV-Ray functions
 * ----------------------------------------------------------------------
 */


/* Place camera */
void 
camera(double dist,double aspect, FILE *out)
{
  if (dist==0) dist=6;
  fprintf(out,"camera {\n"
              "location <0,0,%f>\n",dist);
  fprintf(out,"direction <2,0,0>\n");
  fprintf(out,"right <%f,0,0>",aspect);
  fprintf(out,"look_at <0,0,0>\n");
  fprintf(out,"}\n\n");


}


/* smooth_triangle */
void 
s_triangle(struct po *p1, struct po *p2, struct po *p3, 
	   struct ve *n1, struct ve *n2, struct ve *n3, struct col *c,FILE *out)
{
  /* Is this a corrupt triangle? (Line) */
  if (peq(p1,p2) || peq(p2,p3) || peq(p1,p3)) return;
  
  fprintf(out,"smooth_triangle {\n");
  fprintf(out,"<%-.15f,%-.15f,%-.15f>\n"
	  "<%-.15f,%-.15f,%-.15f>\n",p1->x,p1->y,p1->z,n1->x,n1->y,n1->z);
  fprintf(out,"<%-.15f,%-.15f,%-.15f>\n"
	  "<%-.15f,%-.15f,%-.15f>\n",p2->x,p2->y,p2->z,n2->x,n2->y,n2->z);
  fprintf(out,"<%-.15f,%-.15f,%-.15f>\n"
	  "<%-.15f,%-.15f,%-.15f>\n",p3->x,p3->y,p3->z,n3->x,n3->y,n3->z);

  if (c->r!=NDEF) fprintf(out,"pigment{color rgb <%.5f,%.5f,%.5f>}\n",
			    c->r,c->g,c->b);
  fprintf(out,"}\n");


}


void 
triangle(struct po *p1, struct po *p2, struct po *p3, struct col *c, FILE *out) 
{
  /* Is this a corrupt triangle? (Line) */
  if (peq(p1,p2) || peq(p2,p3) || peq(p1,p3)) return;
  
  fprintf(out,"triangle {\n");
  fprintf(out,"<%-.15f,%-.15f,%-.15f>\n",p1->x,p1->y,p1->z);
  fprintf(out,"<%-.15f,%-.15f,%-.15f>\n",p2->x,p2->y,p2->z);
  fprintf(out,"<%-.15f,%-.15f,%-.15f>\n",p3->x,p3->y,p3->z);
  if (c->r!=NDEF) fprintf(out,"pigment{color rgb <%.5f,%.5f,%.5f>}\n",
			    c->r,c->g,c->b);
  fprintf(out,"}\n");
}

void 
cylinder(struct po *p1, struct po *p2, double r, FILE *out)
{
  /* Is this a corrupt cylinder? (Line) */
  if (peq(p1,p2) || r<=0) return;

  fprintf(out,"cylinder {\n");
  fprintf(out,"<%-.15f,%-.15f,%-.15f>\n",p1->x,p1->y,p1->z);
  fprintf(out,"<%-.15f,%-.15f,%-.15f>,%-.15f\n",p2->x,p2->y,p2->z,r);
  fprintf(out,"}\n");
}

void 
sphere(struct po *p, double r, FILE *out) 
{
  /* Is this a corrupt sphere? (Line) */
  if (r<=0) return;
  fprintf(out,"sphere {\n");
  fprintf(out,"<%-.15f,%-.15f,%-.15f>,%-.15f\n",p->x,p->y,p->z,r);
  fprintf(out,"}\n");
}















